#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Functions for creating VGG-like models (including VGG-LSTM)"""

import keras as ks
from keras.models import Model
from keras.layers import Input, Dense, Dropout, Activation, Flatten, Convolution3D, BatchNormalization, MaxPooling3D,\
                         Convolution2D, MaxPooling2D, TimeDistributed, CuDNNLSTM, Concatenate, Lambda
from keras import backend as K
from keras.regularizers import l2

from scraps.old_model_scripts.file_dump import get_dimensions_encoding

# ------------- VGG-like model -------------#


def decode_input_dimensions_vgg(n_bins, batchsize, swap_col, str_ident =''):
    """
    Returns the general dimension (2D/3D), the input dimensions (i.e. bs x 11 x 13 x 18 x channels=1 for 3D)
    and appropriate max_pool_sizes depending on the projection type.

    Parameters
    ----------
    n_bins : dict(tuple(int))
        Number of bins (x,y,z,t) of the data. Can contain multiple n_bins tuples for different inputs.
    batchsize : int
        Batchsize that is used for the training / inferencing of the cnn.
    swap_col : None/str
        For 4D data input (3.5D models). Specifies, if the channels of the 3.5D net should be swapped.
    str_ident : str
        Optional string identifier that gets appended to the modelname.

    Returns
    -------
    dim : int
        Dimension of the network (2D/3D/4D).
    input_dim : tuple
        Dimensions tuple for 2D, 3D or 4D data (ints).
    max_pool_sizes : dict/dict(dict)
        Dict that specifies the Pooling locations and properties for a cnn
        Key: After which convolutional layer a pooling layer be inserted (counts from layer zero!)
        Item: Specifies the strides.

    """
    # TODO Change this to actually use the dict keywords!
    n_bins = list(n_bins.values())
    if n_bins[0].count(1) == 1:  # 3d case
        dim = 3
        input_dim = get_dimensions_encoding(n_bins[0], batchsize)  # includes batchsize

        if n_bins[0][3] == 1:
            print('Using a VGG-like 3D CNN with XYZ projection')
            max_pool_sizes = {3: (2, 2, 2), 7: (2, 2, 2)}

        elif n_bins[0][0] == 1:
            print('Using a VGG-like 3D CNN with YZT projection')
            max_pool_sizes = {1: (1, 1, 2), 3: (2, 2, 2), 7: (2, 2, 2)}

        else:
            raise IndexError('No suitable 3D projection found in decode_input_dimensions().'
                             'Please add the projection type with the pooling dict to the function.')

    elif n_bins[0].count(1) == 0:  # 4d case, 3.5D
        dim = 3
        if len(n_bins) > 1:

            if swap_col == 'xyz-t_and_yzt-x' and 'tight-1_tight-2' not in str_ident:
                max_pool_sizes = {'net_1': {3: (2, 2, 2), 7: (2, 2, 2)},  # only used if training from scratch
                                  'net_2': {1: (1, 1, 2), 3: (2, 2, 2), 7: (2, 2, 2)}}
                input_dim = get_dimensions_encoding(n_bins[0], batchsize)  # includes batchsize
                input_dim = [(input_dim[0], input_dim[1], input_dim[2], input_dim[3], input_dim[4]),  # xyz-t
                             (input_dim[0], input_dim[2], input_dim[3], input_dim[4], input_dim[1])]  # yzt-x

            elif swap_col == 'xyz-t_and_yzt-x' and 'multi_input_single_train_tight-1_tight-2' in str_ident:
                max_pool_sizes = {'net_1': {3: (2, 2, 2), 7: (2, 2, 2)},  # only used if training from scratch
                                  'net_2': {1: (1, 1, 2), 3: (2, 2, 2), 7: (2, 2, 2)},
                                  'net_3': {3: (2, 2, 2), 7: (2, 2, 2)},
                                  'net_4': {1: (1, 1, 2), 3: (2, 2, 2), 7: (2, 2, 2)}}
                input_dim = get_dimensions_encoding(n_bins[0], batchsize)  # includes batchsize
                input_dim = [(input_dim[0], input_dim[1], input_dim[2], input_dim[3], input_dim[4]),  # xyz-t
                             (input_dim[0], input_dim[2], input_dim[3], input_dim[4], input_dim[1]),  # yzt-x
                             (input_dim[0], input_dim[1], input_dim[2], input_dim[3], input_dim[4]),  # xyz-t
                             (input_dim[0], input_dim[2], input_dim[3], input_dim[4], input_dim[1])]  # yzt-x

            elif swap_col == 'xyz-t_and_xyz-c_single_input':
                max_pool_sizes = {5: (2, 2, 2), 9: (2, 2, 2)}  # 2 more layers ; same as swap_col=None for len(n_bins)=1
                input_dim = get_dimensions_encoding(n_bins[0], batchsize)
                input_dim = (input_dim[0], input_dim[1], input_dim[2], input_dim[3], input_dim[4] + 31)

            elif swap_col is None:
                max_pool_sizes = {'net_1': {5: (2, 2, 2), 9: (2, 2, 2)},
                                  'net_2': {5: (2, 2, 2), 9: (2, 2, 2)}}
                input_dim = get_dimensions_encoding(n_bins[0], batchsize)  # includes batchsize
                input_dim = [(input_dim[0], input_dim[1], input_dim[2], input_dim[3], input_dim[4]),  # xyz-t tight-1
                             (input_dim[0], input_dim[1], input_dim[2], input_dim[3], input_dim[4])]  # xyz-t tight-2

            else:
                raise IOError('3.5D projection types with len(n_bins) > 1 other than "yzt-x_all-t_and_yzt-x_tight-1-t" are not yet supported.')

        else:
            input_dim = get_dimensions_encoding(n_bins[0], batchsize)  # includes batchsize
            if swap_col is None:
                print('Using a VGG-like 3.5D CNN with XYZ data and T/C channel information.')
                # max_pool_sizes = {3: (2, 2, 2), 7: (2, 2, 2)}
                max_pool_sizes = {5: (2, 2, 2), 9: (2, 2, 2)} # 2 more layers
                # max_pool_sizes = {7: (2, 2, 2), 11: (2, 2, 2)}  # 4 more layers

            elif swap_col == 'yzt-x':
                print('Using a VGG-like 3.5D CNN with YZT data and X channel information.')
                # max_pool_sizes = {1: (1, 1, 2), 3: (2, 2, 2), 7: (2, 2, 2)}
                max_pool_sizes = {2: (1, 1, 2), 5: (2, 2, 2), 9: (2, 2, 2)} # 2 more layers
                input_dim = (input_dim[0], input_dim[2], input_dim[3], input_dim[4], input_dim[1])  # [bs,y,z,t,x]

            elif swap_col == 'xyz-t_and_yzt-x':
                max_pool_sizes = {'net_1': {3: (2, 2, 2), 7: (2, 2, 2)},
                                  'net_2': {1: (1, 1, 2), 3: (2, 2, 2), 7: (2, 2, 2)}}
                input_dim = [(input_dim[0], input_dim[1], input_dim[2], input_dim[3], input_dim[4]),  # xyz-t
                             (input_dim[0], input_dim[2], input_dim[3], input_dim[4], input_dim[1])]  # yzt-x

            elif swap_col == 'conv_lstm':  # TODO fix whole function to make it more general and not only 3.5D stuff
                max_pool_sizes = {1: (2, 2, 2), 5: (2, 2, 2)}
                input_dim = (input_dim[0], input_dim[4], input_dim[1], input_dim[2], input_dim[3], 1)  # t-xyz

            else:
                raise IOError('3.5D projection types other than XYZ-T and YZT-X are not yet supported.'
                              'Please add the max_pool_sizes dict in the function by yourself.')

    else:
        raise IOError('Data types other than 3D or 4D (3.5D actually) are not yet supported. '
                      'Please specify a 3D or 4D n_bins tuple.')

    return dim, input_dim, max_pool_sizes


def create_vgg_like_model(n_bins, class_type, n_filters=None, dropout=0, k_size=3, swap_col=None,
                          activation='relu', kernel_reg=None):
    """
    Returns a VGG-like model (stacked conv. layers) with MaxPooling and Dropout if wished.

    The number of convolutional layers can be controlled with the n_filters parameter:
    n_conv_layers = len(n_filters)

    Parameters
    ----------
    n_bins : dict(tuple(int))
        Number of bins (x,y,z,t) of the data. Can contain multiple n_bins tuples for different inputs.
    class_type : str
        Declares the number of output classes / regression variables and a string identifier to specify the exact output classes.
    n_filters : tuple
        Number of filters for each conv. layer. len(n_filters)=n_conv_layer.
    dropout : float
        Adds dropout if >0.
    k_size : int
        Kernel size which is used for all dimensions.
    swap_col : None/str
        For 4D data input (3.5D models). Specifies, if the channels of the 3.5D net should be swapped.
    activation : str
        Type of activation function that should be used for the net. E.g. 'linear', 'relu', 'elu', 'selu'.
    kernel_reg : None/str
        If L2 regularization with 1e-4 should be employed. 'l2' to enable the regularization.

    Returns
    -------
    model : ks.models.Model
        A VGG-like Keras nn instance.

    """
    if n_filters is None: n_filters = (64, 64, 64, 64, 64, 128, 128, 128)
    if kernel_reg is 'l2': kernel_reg = l2(0.0001)

    # TODO Batchsize has to be given to decode_input_dimensions_vgg, but is not used for constructing the model.
    # For now: Just use some random value.
    batchsize = 64

    dim, input_dim, max_pool_sizes = decode_input_dimensions_vgg(n_bins, batchsize, swap_col)

    input_layer = Input(shape=input_dim[1:], name=swap_col, dtype=K.floatx())  # input_layer
    x = conv_block(input_layer, dim, n_filters[0], k_size=k_size, dropout=dropout, max_pooling=max_pool_sizes.get(0), activation=activation, kernel_reg=kernel_reg)

    for i in range(1, len(n_filters)):
        x = conv_block(x, dim, n_filters[i], k_size=k_size, dropout=dropout, max_pooling=max_pool_sizes.get(i), activation=activation, kernel_reg=kernel_reg)

    conv_output_flat = Flatten()(x)
    outputs = add_dense_layers_to_cnn(conv_output_flat, class_type, dropout=dropout, activation=activation, kernel_reg=kernel_reg)

    model = Model(inputs=input_layer, outputs=outputs)

    return model


def conv_block(ip, dim, n_filters, k_size=3, dropout=0, max_pooling=None, activation='relu', kernel_reg = None):
    """
    2D/3D Convolutional block followed by BatchNorm and Activation with optional MaxPooling or Dropout.

    C-B-A-(MP)-(D)

    Parameters
    ----------
    ip : ? # TODO
        Keras functional layer instance that is used as the starting point of this convolutional block.
    dim : int
        Specifies the dimension of the convolutional block, 2D/3D.
    n_filters : int
        Number of filters used for the convolutional layer.
    k_size : int
        Kernel size which is used for all three dimensions.
    dropout : float
        Adds a dropout layer if the value is greater than 0.
    max_pooling : None/tuple
        Specifies if a MaxPooling layer should be added. e.g. (1,1,2) -> strides for a 3D conv block.
    activation : str
        Type of activation function that should be used. E.g. 'linear', 'relu', 'elu', 'selu'.
    kernel_reg : None/str
        If L2 regularization with 1e-4 should be employed. 'l2' to enable the regularization.

    Returns
    -------
    x :
        Resulting output tensor (model).

    """
    if dim not in (2,3): raise ValueError('dim must be equal to 2 or 3.')
    convolution_nd = Convolution2D if dim==2 else Convolution3D
    max_pooling_nd = MaxPooling2D if dim==2 else MaxPooling3D

    channel_axis = 1 if K.image_data_format() == "channels_first" else -1

    x = convolution_nd(n_filters, (k_size,) * dim, padding='same', kernel_initializer='he_normal', use_bias=False, kernel_regularizer=kernel_reg)(ip)

    x = BatchNormalization(axis=channel_axis)(x)
    x = Activation(activation)(x)

    if max_pooling is not None: x = max_pooling_nd(strides=max_pooling, padding='valid')(x)
    if dropout > 0.0: x = Dropout(dropout)(x)

    return x


def add_dense_layers_to_cnn(conv_output_flat, class_type, dropout=0, activation='relu', kernel_reg=None):
    """
    Appends dense layers to the convolutional layers of a cnn.

    Parameters
    ----------
    conv_output_flat : ? # TODO
        The Keras layer instance after the Flatten() layer.
    class_type : str
        Declares the number of output classes / regression variables and a string identifier to specify the exact output classes.
    dropout : float
        Adds a dropout layer if the value is greater than 0.
    activation : str
        Type of activation function that should be used. E.g. 'linear', 'relu', 'elu', 'selu'.
    kernel_reg : None/str
        If L2 regularization with 1e-4 should be employed. 'l2' to enable the regularization.

    Returns
    -------
    outputs : list
        List of outputs for the cnn.

    """

    x = Dense(128, kernel_initializer='he_normal', kernel_regularizer=kernel_reg, activation=activation)(conv_output_flat)
    if dropout > 0.0: x = Dropout(dropout)(x)
    x = Dense(32, kernel_initializer='he_normal', kernel_regularizer=kernel_reg, activation=activation)(x)

    outputs = []

    if class_type == 'ts_classifier':  # categorical problem
        x = Dense(2, activation='softmax', kernel_initializer='he_normal', name='ts_output')(x)
        outputs.append(x)

    elif class_type == 'bg_classifier':  # categorical problem
        x = Dense(3, activation='softmax', kernel_initializer='he_normal', name='bg_output')(x)
        outputs.append(x)

    elif class_type == 'bg_classifier_2_class':  # categorical problem
        x = Dense(2, activation='softmax', kernel_initializer='he_normal', name='bg_output')(x)
        outputs.append(x)

    else:  # regression case, one output for each regression label

        if class_type == 'energy_dir_bjorken-y_errors':
            label_names = ('e', 'dir_x', 'dir_y', 'dir_z', 'by')
            for name in label_names:
                output_label = Dense(1, name=name)(x)
                outputs.append(output_label)

            # build second dense network for errors
            conv_output_flat_stop_grad = Lambda(lambda a: K.stop_gradient(a))(conv_output_flat)
            x_err = Dense(128, kernel_initializer='he_normal', kernel_regularizer=kernel_reg, activation=activation)(
                conv_output_flat_stop_grad)
            if dropout > 0.0: x_err = Dropout(dropout)(x_err)
            x_err = Dense(64, kernel_initializer='he_normal', kernel_regularizer=kernel_reg, activation=activation)(
                x_err)
            if dropout > 0.0: x_err = Dropout(dropout)(x_err)
            x_err = Dense(32, kernel_initializer='he_normal', kernel_regularizer=kernel_reg, activation=activation)(
                x_err)

            for i, name in enumerate(label_names):
                output_label_error = Dense(1, activation='linear', name=name + '_err_temp')(x_err)
                # second stop_gradient through the concat layer into the dense_nn of the labels is in the err_loss function
                output_label_merged = Concatenate(name=name + '_err')([outputs[i], output_label_error])
                outputs.append(output_label_merged)

        elif class_type == 'energy_dir_bjorken-y_vtx_errors':
            label_names = ('e', 'dx', 'dy', 'dz', 'by', 'vx', 'vy', 'vz', 'vt')
            for name in label_names:
                output_label = Dense(1, name=name)(x)
                outputs.append(output_label)

            # build second dense network for errors
            conv_output_flat_stop_grad = Lambda(lambda a: K.stop_gradient(a))(conv_output_flat)
            x_err = Dense(128, kernel_initializer='he_normal', kernel_regularizer=kernel_reg, activation=activation)(conv_output_flat_stop_grad)
            if dropout > 0.0: x_err = Dropout(dropout)(x_err)
            x_err = Dense(64, kernel_initializer='he_normal', kernel_regularizer=kernel_reg, activation=activation)(x_err)
            if dropout > 0.0: x_err = Dropout(dropout)(x_err)
            x_err = Dense(32, kernel_initializer='he_normal', kernel_regularizer=kernel_reg, activation=activation)(x_err)

            for i, name in enumerate(label_names):
                output_label_error = Dense(1, activation='linear', name=name + '_err_temp')(x_err)
                # second stop_gradient through the concat layer into the dense_nn of the labels is in the err_loss function
                output_label_merged = Concatenate(name=name + '_err')([outputs[i], output_label_error])
                outputs.append(output_label_merged)

        else:
            raise ValueError(class_type, "is not a known class_type!")

    return outputs


def create_vgg_like_model_double_input(n_bins, batchsize, nb_classes=2, n_filters=None, dropout=0, k_size=3, swap_4d_channels=None, activation='relu'):
    """
    Returns a double input, VGG-like model (stacked conv. layers) with MaxPooling and Dropout if wished for classification.

    The two single VGG networks are concatenated after the last flatten layers.
    The number of convolutional layers can be controlled with the n_filters parameter:
    n_conv_layers = len(n_filters)

    Parameters
    ----------
    n_bins : list(tuple(int))
        Number of bins (x,y,z,t) of the data. Can contain multiple n_bins tuples.
    batchsize : int
        Batchsize that is used for the training / inferencing of the cnn.
    nb_classes : int
        Number of output classes.
    n_filters : tuple
        Number of filters for each conv. layer. len(n_filters)=n_conv_layer.
    dropout : float
        Adds dropout if >0.
    k_size : int
        Kernel size which is used for all dimensions.
    swap_4d_channels : None/str
        For 4D data input (3.5D models). Specifies, if the channels of the 3.5D net should be swapped.
    activation : str
        Type of activation function that should be used for the net. E.g. 'linear', 'relu', 'elu', 'selu'.

    Returns
    -------
    model : ks.models.Model
        A VGG-like, double input Keras nn instance.

    """
    if n_filters is None: n_filters = (64,64,64,64,64,128,128,128)

    dim, input_dim, max_pool_sizes = decode_input_dimensions_vgg(n_bins, batchsize, swap_4d_channels)

    input_layer_1 = Input(shape=input_dim[0][1:], dtype=K.floatx())  # input_layer 1
    input_layer_2 = Input(shape=input_dim[1][1:], dtype=K.floatx())  # input_layer 2

    # Net 1 convs
    x_1 = conv_block(input_layer_1, dim, n_filters[0], k_size=k_size, dropout=dropout, max_pooling=max_pool_sizes['net_1'].get(0), activation=activation)

    for i in range(1, len(n_filters)):
        x_1 = conv_block(x_1, dim, n_filters[i], k_size=k_size, dropout=dropout, max_pooling=max_pool_sizes['net_1'].get(i), activation=activation)

    # Net 2 convs
    x_2 = conv_block(input_layer_2, dim, n_filters[0], k_size=k_size, dropout=dropout, max_pooling=max_pool_sizes['net_2'].get(0), activation=activation)

    for i in range(1, len(n_filters)):
        x_2 = conv_block(x_2, dim, n_filters[i], k_size=k_size, dropout=dropout, max_pooling=max_pool_sizes['net_2'].get(i), activation=activation)

    # flatten both nets
    x_1, x_2 = Flatten()(x_1), Flatten()(x_2)

    # concatenate both nets
    x = ks.layers.concatenate([x_1, x_2])

    x = Dense(128, activation=activation, kernel_initializer='he_normal')(x)
    if dropout > 0.0: x = Dropout(dropout)(x)
    x = Dense(16, activation=activation, kernel_initializer='he_normal')(x)

    x = Dense(nb_classes, activation='softmax', kernel_initializer='he_normal')(x)

    model = Model(inputs=[input_layer_1, input_layer_2], outputs=x)

    return model


def create_vgg_like_model_multi_input_from_single_nns(n_bins, str_ident, dropout=(0, 0.2), swap_4d_channels=None, activation='relu'):
    """
    Returns a double input, VGG-like model (stacked conv. layers) with MaxPooling and Dropout if wished.

    The two single VGG networks are concatenated after the last flatten layers.

    Parameters
    ----------
    n_bins : list(tuple(int))
        Number of bins (x,y,z,t) of the data. Can contain multiple n_bins tuples.
    str_ident : str
        Optional string identifier that gets appended to the modelname.
    nb_classes : int
        Number of output classes.
    dropout : tuple(float, float)
        Adds dropout if >0.
    swap_4d_channels : None/str
        For 4D data input (3.5D models). Specifies, if the channels of the 3.5D net should be swapped.
    activation : str
        Type of activation function that should be used for the net. E.g. 'linear', 'relu', 'elu', 'selu'.

    Returns
    -------
    model : ks.models.Model
        A VGG-like, double input Keras nn instance, with pretrained conv layers.

    """
    # TODO Batchsize has to be given to decode_input_dimensions_vgg, but is not used for constructing the model.
    # For now: Just use some random value.
    batchsize=64
    nb_classes = 2

    dim, input_dim, max_pool_sizes = decode_input_dimensions_vgg(n_bins, batchsize, swap_4d_channels, str_ident=str_ident)
    trained_model_paths = {}
    if swap_4d_channels is None: swap_4d_channels = ''

    if 'xyz-t_and_yzt-x' + 'multi_input_single_train_tight-1' in swap_4d_channels + str_ident and 'multi_input_single_train_tight-1_tight-2' not in str_ident:
        trained_model_paths[0] = 'models/trained/trained_model_VGG_4d_xyz-t_muon-CC_to_elec-CC_xyz-t_tight-1_w-geo-fix_bs64_2-more-layers_epoch_32_file_1.h5'  # xyz-t, timecut tight_1, with geo fix, 2 more layers
        trained_model_paths[1] = 'models/trained/trained_model_VGG_4d_yzt-x_muon-CC_to_elec-CC_tight-1_w-geo-fix_bs64_dp0.1_2-more-layers_epoch_30_file_1.h5'  # yzt-x, timecut tight-1, with geo fix, 2 more layers

    elif swap_4d_channels is '': # xyz-t tight-1 and xyz-t tight-2
        trained_model_paths[0] = 'models/trained/trained_model_VGG_4d_xyz-t_track-shower_lp_tight-1_bs64_dp0.1_pad_same_epoch_15_file_1.h5'  # lp, xyz-t, timecut tight_1, fully trained epoch 16, file 1
        trained_model_paths[1] = 'models/trained/trained_model_VGG_4d_xyz-t_track-shower_lp_tight-2_bs64_dp0.1_padsame_epoch_12_file_1.h5'  # lp, xyz-t, timecut tight-2, fully trained epoch 13, file 1

    elif 'xyz-t_and_yzt-x' + 'multi_input_single_train_tight-1_tight-2' in swap_4d_channels + str_ident:
        trained_model_paths[0] = 'models/trained/trained_model_VGG_4d_xyz-t_track-shower_lp_tight-1_bs64_dp0.1_pad_same_epoch_16_file_1.h5'  # lp, xyz-t, timecut tight-1, fully trained epoch 16, file 1
        trained_model_paths[2] = 'models/trained/trained_model_VGG_4d_xyz-t_track-shower_lp_tight-2_bs64_dp0.1_padsame_epoch_13_file_1.h5'  # lp, xyz-t, timecut tight-2, fully trained epoch 13, file 1
        trained_model_paths[1] = 'models/trained/trained_model_VGG_4d_yzt-x_track-shower_lp_tight-1_bs64_dp0.1_padsame_epoch_10_file_3.h5' # lp, yzt-x, timecut tight-1, fully trained epoch 10, file 3
        trained_model_paths[3] = 'models/trained/trained_model_VGG_4d_yzt-x_track-shower_lp_tight-2_bs64_dp0.1_padsame_epoch_11_file_2.h5' # lp, yzt-x, timecut tight-2, fully trained epoch 11, file 2

    else:
        raise ValueError('The double input combination specified in "swap_4d_channels" is not known, check the function for what is available.')

    n_inputs = len(trained_model_paths)
    trained_models = {}
    input_layers = {}
    layer_numbers = {}
    x = {}

    for i in range(n_inputs):
        trained_models[i] = ks.models.load_model(trained_model_paths[i])

        input_layers[i] = Input(shape=input_dim[i][1:], name='input_net_' + str(i+1), dtype=K.floatx())
        layer_numbers[i] = {'conv': 1, 'batch_norm': 1, 'activation': 1, 'max_pooling': 1, 'dropout': 1}

        x[i] = create_layer_from_config(input_layers[i], trained_models[i].layers[1], layer_numbers[i], trainable=False, net=str(i+1))
        for trained_layer in trained_models[i].layers[2:]:
            if 'flatten' in trained_layer.name: break  # we don't want to get anything after the flatten layer
            x[i] = create_layer_from_config(x[i], trained_layer, layer_numbers[i], trainable=False, net=str(i+1), dropout=dropout[0])

        x[i] = Flatten()(x[i])

    x = ks.layers.concatenate([x[i] for i in x])

    x = Dense(128, activation=activation, kernel_initializer='he_normal')(x)
    x = Dropout(dropout[1])(x)
    x = Dense(32, activation=activation, kernel_initializer='he_normal')(x)

    x = Dense(nb_classes, activation='softmax', kernel_initializer='he_normal')(x)

    model = Model(inputs=[input_layers[i] for i in input_layers], outputs=x)
    set_layer_weights(model, trained_models) # set weights

    for layer in model.layers: # freeze trainable batch_norm weights, but not running mean and variance
        if 'batch_norm' in layer.name:
            layer.stateful = True

    return model


def create_layer_from_config(x, trained_layer, layer_numbers, trainable=False, net='', dropout=0):
    """
    Creates a new Keras nn layer from the config of an already existing layer.

    Changes the 'trainable' flag of the new layer to false and optionally udates the dropout rate.
    Adds a layer name based on the layer_numbers dict.

    Parameters
    ----------
    x : ? # TODO
        Keras functional model api instance. E.g. TF tensors.
    trained_layer : ks.layer
        Keras layer instance that is already trained.
    layer_numbers : dict
        Dictionary for the different layer types to keep track of the layer_number in the layer names.
    trainable : bool
        Flag to set the <trainable> attribute of the new layer.
    net : str
        Additional string that is added to the layer name. E.g. 'net_2' if a double input model is used.
    dropout : float
        optional, dropout rate of the new layer

    Returns
    -------
    x : ?
        Keras functional model api instance. E.g. TF tensors. Contains a new layer now!

    """
    if 'conv' in trained_layer.name:
        layer, name = Convolution3D, 'conv'
    elif 'batch_norm' in trained_layer.name:
        layer, name = BatchNormalization, 'batch_norm'
    elif 'activation' in trained_layer.name:
        layer, name = Activation, 'activation'
    elif 'pooling' in trained_layer.name:
        layer, name = MaxPooling3D, 'max_pooling'
    elif 'dropout' in trained_layer.name:
        layer, name = Dropout, 'dropout'
    elif 'dense' in trained_layer.name:
        layer, name = Dense, 'dense'
    else:
        return x # if 'input' or 'flatten'

    config = trained_layer.get_config()
    config.update({'trainable': trainable}) # for freezing the layer if wanted
    if name == 'dropout': config.update({'rate': dropout})

    if net == '':
        new_layer_name = name + '_' + str(layer_numbers[name])
    else:
        new_layer_name = name + '_' + str(layer_numbers[name]) + '_net_' + net

    layer_numbers[name] = layer_numbers[name] + 1
    config.update({'name': new_layer_name})

    x = layer.from_config(config)(x)

    return x


def set_layer_weights(model, trained_models):
    """
    Sets the weights of a double input model (until the first flatten layer) based on two pretrained models.

    Parameters
    ----------
    model : ks.models.Model
        Keras model instance.
    trained_models : dict
        Dict that contains references to the Keras model instances of the already pretrained models.

    """
    skip_layers = ['dropout', 'input', 'dense', 'flatten', 'max_pooling', 'activation', 'concatenate']
    n_models = len(trained_models)
    trained_layers_w_weights = {}

    for i in range(n_models):
        trained_layers_w_weights[i] = [layer for layer in trained_models[i].layers if 'conv' in layer.name or 'batch_normalization' in layer.name]  # still ordered

        j = -1
        for layer in model.layers:
            if 'net_' + str(i+1) not in layer.name: continue

            skip = False  # workaround of hell...
            for skip_layer_str in skip_layers:
                if skip_layer_str in layer.name:
                    skip = True
            if skip: continue

            j += 1
            layer.set_weights(trained_layers_w_weights[i][j].get_weights())


def change_dropout_rate_for_multi_input_model(n_bins, batchsize, trained_model, dropout=(0.1, 0.1), trainable=(True, True), swap_4d_channels=None):
    """
    Function that rebuilds a keras model and modifies its dropout rate. Workaround, till layer.rate is fixed to work with Dropout layers.

    Parameters
    ----------
    n_bins : list(tuple(int))
        Number of bins (x,y,z,t) of the data. Can contain multiple n_bins tuples.
    batchsize : int
        Batchsize that is used for the training / inferencing of the cnn.
    trained_model : ks.models.Model
        Trained Keras model, upon which the dropout rate should be changed.
    dropout : (float, float)
        Adds dropout if > 0. First value for the conv block, second value for the dense.
    trainable : (bool, bool)
        Sets the trainable flag for the conv block layers and for the dense layers.
    swap_4d_channels : None/str
        For 4D data input (3.5D models). Specifies, if the channels of the 3.5D net should be swapped.
        Only used to decode the input_dim.

    Returns
    -------
    model : ks.models.Model
        Keras VGG-like model based on the trained_model with modified dropout layers.

    """
    dim, input_dim, max_pool_sizes = decode_input_dimensions_vgg(n_bins, batchsize, swap_4d_channels)

    # rebuild trained_model based on it's layer config
    input_layer_net_1 = Input(shape=input_dim[0][1:], name='input_net_1', dtype=K.floatx())  # have to do that manually
    input_layer_net_2 = Input(shape=input_dim[1][1:], name='input_net_2', dtype=K.floatx())

    layer_numbers_net_1 = {'conv': 1, 'batch_norm': 1, 'activation': 1, 'max_pooling': 1, 'dropout': 1}
    layer_numbers_net_2 = {'conv': 1, 'batch_norm': 1, 'activation': 1, 'max_pooling': 1, 'dropout': 1}
    layer_names = [layer.name for layer in trained_model.layers]

    x_1, x_2 = input_layer_net_1, input_layer_net_2 # needed for creating the trained_model later on
    for trained_layer in trained_model.layers[2:]:
        if 'net_1' in trained_layer.name:
            x_1 = create_layer_from_config(x_1, trained_layer, layer_numbers_net_1, trainable=trainable[0], net='1', dropout=dropout[0])

        elif 'net_2' in trained_layer.name:
            x_2 = create_layer_from_config(x_2, trained_layer, layer_numbers_net_2, trainable=trainable[0], net='2', dropout=dropout[0])

        else: break

    # flatten both nets and concatenate them
    x_1, x_2 = Flatten()(x_1), Flatten()(x_2)
    x = ks.layers.concatenate([x_1, x_2])

    # rebuild dense layers
    layer_index_first_dense = layer_names.index('concatenate_1') + 1
    layer_numbers_fc = {'dense': 1, 'dropout': 1}

    for trained_layer in trained_model.layers[layer_index_first_dense:]:
        x = create_layer_from_config(x, trained_layer, layer_numbers_fc, trainable=trainable[1], dropout=dropout[1])

    model = Model(inputs=[input_layer_net_1, input_layer_net_2], outputs=x)

    set_layer_weights_from_single_trained_model(model, trained_model) # set weights

    return model


def set_layer_weights_from_single_trained_model(model, trained_model):
    """
    Sets the weights of a Keras model based on an already trained trained_model.

    Parameters
    ----------
    model : ks.models.Model
        Keras model instance withought weights.
    trained_model : ks.models.Model
        Pretrained Keras model used to set the weights for the new model.

    """
    skip_layers = ['dropout', 'input', 'flatten', 'max_pooling', 'activation', 'concatenate']

    i = -1
    for layer in model.layers:
        i += 1

        skip = False # workaround of hell...
        for skip_layer_str in skip_layers:
            if skip_layer_str in layer.name:
                skip = True
        if skip: continue

        layer.set_weights(trained_model.layers[i].get_weights())


def create_convolutional_lstm(n_bins, batchsize, nb_classes=2, n_filters=None, dropout=0, k_size=3, activation='relu', kernel_reg=None):
    """
    Returns a VGG-like, convolutional LSTM model (stacked conv. layers + LSTM) with MaxPooling and Dropout if wished.
    Only working for classification as of now!

    The number of convolutional layers can be controlled with the n_filters parameter:
    n_conv_layers = len(n_filters)

    Parameters
    ----------
    n_bins : list(tuple(int))
        Number of bins (x,y,z,t) of the data. Can contain multiple n_bins tuples.
    batchsize : int
        Batchsize that is used for the training / inferencing of the cnn.
    nb_classes : int
        Number of output classes.
    n_filters : None/tuple
        Number of filters for each conv. layer. len(n_filters)=n_conv_layer.
    dropout : float
        Adds dropout if >0.
    k_size : int
        Kernel size which is used for all dimensions.
    activation : str
        Type of activation function that should be used for the net. E.g. 'linear', 'relu', 'elu', 'selu'.
    kernel_reg : None/str
        If L2 regularization with 1e-4 should be employed. 'l2' to enable the regularization.

    Returns
    -------
    model : ks.models.Model
        Keras conv lstm model.

    """
    if n_filters is None: n_filters = (32, 32, 64, 64, 64, 64, 128)
    if kernel_reg is 'l2': kernel_reg = l2(0.0001)

    dim, input_dim, max_pool_sizes = decode_input_dimensions_vgg(n_bins, batchsize, 'conv_lstm') # TODO fix input dim

    input_layer = Input(shape=input_dim[1:], dtype=K.floatx())  # input_layer
    x = conv_block_time_distributed(input_layer, n_filters[0], k_size=k_size, dropout=dropout, max_pooling=max_pool_sizes.get(0), activation=activation, kernel_reg=kernel_reg)

    for i in range(1, len(n_filters)):
        x = conv_block_time_distributed(x, n_filters[i], k_size=k_size, dropout=dropout, max_pooling=max_pool_sizes.get(i), activation=activation, kernel_reg=kernel_reg)

    x = TimeDistributed(Flatten())(x)

    x = CuDNNLSTM(768)(x)

    if dropout > 0.0: x = Dropout(dropout)(x)

    x = Dense(64, kernel_initializer='he_normal', kernel_regularizer=kernel_reg)(x)
    x = Activation(activation)(x)
    if dropout > 0.0: x = Dropout(dropout)(x)
    x = Dense(16, kernel_initializer='he_normal', kernel_regularizer=kernel_reg)(x)
    x = Activation(activation)(x)

    x = Dense(nb_classes, activation='softmax', kernel_initializer='he_normal')(x)

    model = Model(inputs=input_layer, outputs=x)

    return model


def conv_block_time_distributed(ip, n_filters, k_size=3, dropout=0, max_pooling=None, activation='relu', kernel_reg = None):
    """
    Time distributed 2D/3D Convolutional block followed by BatchNorm and Activation with optional MaxPooling or Dropout.

    C-B-A-(MP)-(D)

    Parameters
    ----------
    ip :
        Keras functional layer instance that is used as the starting point of this convolutional block.
    n_filters : int
        Number of filters used for the convolution.
    k_size : int
        Kernel size which is used for all three dimensions.
    dropout : float
        Adds a dropout layer if value is greater than 0.
    max_pooling : None/tuple
        Specifies if a MaxPooling layer should be added. e.g. (1,1,2) -> strides for a 3D conv block.
    activation : str
        Type of activation function that should be used. E.g. 'linear', 'relu', 'elu', 'selu'.
    kernel_reg : None/str
        If L2 regularization with 1e-4 should be employed. 'l2' to enable the regularization.

    Returns
    -------
    x :
        Resulting output tensor (model).

    """
    x = TimeDistributed(Convolution3D(n_filters, (k_size,) * 3, padding='same', kernel_initializer='he_normal', use_bias=False, kernel_regularizer=kernel_reg))(ip)

    x = TimeDistributed(BatchNormalization(axis=-1))(x)
    x = Activation(activation)(x)

    if max_pooling is not None: x = TimeDistributed(MaxPooling3D(pool_size=max_pooling, padding='valid'))(x)
    if dropout > 0.0: x = TimeDistributed(Dropout(dropout))(x)

    return x


#------------- VGG-like model -------------#












"""
Multi-gpu code for Keras/TF.
From https://github.com/avolkov1/keras_experiments
"""
from keras.legacy import interfaces
from keras import callbacks as cbks


__all__ = ('ModelDataflowMixin',)


class ModelDataflowMixin(object):
    @interfaces.legacy_generator_methods_support
    def fit_dataflow(self, dflow,
                     steps_per_epoch,
                     epochs=1,
                     verbose=1,
                     callbacks=None,
                     validation_data=None,
                     validation_steps=None,
                     class_weight=None,
                     max_q_size=10,
                     workers=1,
                     pickle_safe=False,
                     initial_epoch=0):
        """Fits the model on data yielded batch-by-batch by a Python generator.
        The generator is run in parallel to the model, for efficiency.
        For instance, this allows you to do real-time data augmentation
        on images on CPU in parallel to training your model on GPU.
        # Arguments
            dflow: a dataflow object a-la-carte Tensorpack.
                The output of the generator must be either
                - a tuple (inputs, targets)
                - a tuple (inputs, targets, sample_weights).
                All arrays should contain the same number of samples.
                The generator is expected to loop over its data
                indefinitely. An epoch finishes when `steps_per_epoch`
                samples have been seen by the model.
            steps_per_epoch: Total number of steps (batches of samples)
                to yield from `generator` before declaring one epoch
                finished and starting the next epoch. It should typically
                be equal to the number of unique samples if your dataset
                divided by the batch size.
            epochs: integer, total number of iterations on the data.
            verbose: verbosity mode, 0, 1, or 2.
            callbacks: list of callbacks to be called during training.
            validation_data: this can be either
                - a generator for the validation data
                - a tuple (inputs, targets)
                - a tuple (inputs, targets, sample_weights).
            validation_steps: Only relevant if `validation_data`
                is a generator. Total number of steps (batches of samples)
                to yield from `generator` before stopping.
            class_weight: dictionary mapping class indices to a weight
                for the class.
            max_q_size: maximum size for the generator queue
            workers: maximum number of processes to spin up
                when using process based threading
            pickle_safe: if True, use process based threading.
                Note that because
                this implementation relies on multiprocessing,
                you should not pass
                non picklable arguments to the generator
                as they can't be passed
                easily to children processes.
            initial_epoch: epoch at which to start training
                (useful for resuming a previous training run)
        # Returns
            A `History` object.
        # Example
        ```python
            def generate_arrays_from_file(path):
                while 1:
                    f = open(path)
                    for line in f:
                        # create numpy arrays of input data
                        # and labels, from each line in the file
                        x1, x2, y = process_line(line)
                        yield ({'input_1': x1, 'input_2': x2}, {'output': y})
                    f.close()
            model.fit_generator(generate_arrays_from_file('/my_file.txt'),
                                steps_per_epoch=10000, epochs=10)
        ```
        # Raises
            ValueError: In case the generator yields
                data in an invalid format.
        """
        # wait_time = 0.01  # in seconds
        epoch = initial_epoch

        do_validation = bool(validation_data)
        self._make_train_function()
        if do_validation:
            self._make_test_function()

        # python 2 has 'next', 3 has '__next__'
        # avoid any explicit version checks
        val_gen = (hasattr(validation_data, 'next') or
                   hasattr(validation_data, '__next__'))
        if val_gen and not validation_steps:
            raise ValueError('When using a generator for validation data, '
                             'you must specify a value for '
                             '`validation_steps`.')

        out_labels = self.metrics_names
        callback_metrics = out_labels + ['val_' + n for n in out_labels]

        # prepare callbacks
        self.history = cbks.History()
        callbacks = [cbks.BaseLogger()] + (callbacks or []) + [self.history]
        if verbose:
            callbacks += [cbks.ProgbarLogger(count_mode='steps')]
        callbacks = cbks.CallbackList(callbacks)

        # it's possible to callback a different model than self:
        if hasattr(self, 'callback_model') and self.callback_model:
            callback_model = self.callback_model
        else:
            callback_model = self
        callbacks.set_model(callback_model)
        callbacks.set_params({
            'epochs': epochs,
            'steps': steps_per_epoch,
            'verbose': verbose,
            'do_validation': do_validation,
            'metrics': callback_metrics,
        })
        callbacks.on_train_begin()

        if do_validation and not val_gen:
            if len(validation_data) == 2:
                val_x, val_y = validation_data
                val_sample_weight = None
            elif len(validation_data) == 3:
                val_x, val_y, val_sample_weight = validation_data
            else:
                raise ValueError('validation_data should be a tuple '
                                 '`(val_x, val_y, val_sample_weight)` '
                                 'or `(val_x, val_y)`. Found: ' +
                                 str(validation_data))
            val_x, val_y, val_sample_weights = self._standardize_user_data(
                val_x, val_y, val_sample_weight)
            for cbk in callbacks:
                cbk.validation_data = val_x + [val_y, val_sample_weights]
        # enqueuer = None

        # TODO: Tensorpack does some kind of acceleratn using
        #     QueueInputTrainer, QueueInput, and EnqueueThread. The
        #     implementation below corresponds to SimpleTrainer which
        #     Tensorpack notes as being slow. I still cannot decipher what
        #     exactly is going on in Tensorpack. For the same per-GPU batchsize
        #     the runtime per epoch seems on par. Perhaps with Tensorpack
        #     implementation using Queue+Thread for datafalow the feed_dict
        #     would be faster. The keras fit_generator does use an enqueuer,
        #     but I did not notice performance difference between using
        #     fit_generator or this mixed-in fit_dataflow method.

        try:
            # enqueuer = GeneratorEnqueuer(generator, pickle_safe=pickle_safe)
            # enqueuer.start(max_q_size=max_q_size, workers=workers)

            dflow.reset_state()
            _generator = dflow.get_data()

            callback_model.stop_training = False
            while epoch < epochs:
                callbacks.on_epoch_begin(epoch)
                steps_done = 0
                batch_index = 0
                while steps_done < steps_per_epoch:
                    # generator_output = None
                    generator_output = next(_generator)
                    # while enqueuer.is_running():
                    #     if not enqueuer.queue.empty():
                    #         generator_output = enqueuer.queue.get()
                    #         break
                    #     else:
                    #         time.sleep(wait_time)

                    if not hasattr(generator_output, '__len__'):
                        raise ValueError('output of generator should be '
                                         'a tuple `(x, y, sample_weight)` '
                                         'or `(x, y)`. Found: ' +
                                         str(generator_output))
                    if len(generator_output) == 2:
                        x, y = generator_output
                        sample_weight = None
                    elif len(generator_output) == 3:
                        x, y, sample_weight = generator_output
                    else:
                        raise ValueError('output of generator should be '
                                         'a tuple `(x, y, sample_weight)` '
                                         'or `(x, y)`. Found: ' +
                                         str(generator_output))
                    # build batch logs
                    batch_logs = {}
                    if isinstance(x, list):
                        batch_size = x[0].shape[0]
                    elif isinstance(x, dict):
                        batch_size = list(x.values())[0].shape[0]
                    else:
                        batch_size = x.shape[0]
                    batch_logs['batch'] = batch_index
                    batch_logs['size'] = batch_size
                    callbacks.on_batch_begin(batch_index, batch_logs)

                    outs = self.train_on_batch(x, y,
                                               sample_weight=sample_weight,
                                               class_weight=class_weight)

                    if not isinstance(outs, list):
                        outs = [outs]
                    for l, o in zip(out_labels, outs):
                        batch_logs[l] = o

                    callbacks.on_batch_end(batch_index, batch_logs)

                    # Construct epoch logs.
                    epoch_logs = {}
                    batch_index += 1
                    steps_done += 1

                    # Epoch finished.
                    if steps_done >= steps_per_epoch and do_validation:
                        if val_gen:
                            val_outs = self.evaluate_generator(
                                validation_data,
                                validation_steps,
                                max_q_size=max_q_size,
                                workers=workers,
                                pickle_safe=pickle_safe)
                        else:
                            # No need for try/except because
                            # data has already been validated.
                            val_outs = self.evaluate(
                                val_x, val_y,
                                batch_size=batch_size,
                                sample_weight=val_sample_weights,
                                verbose=0)
                        if not isinstance(val_outs, list):
                            val_outs = [val_outs]
                        # Same labels assumed.
                        for l, o in zip(out_labels, val_outs):
                            epoch_logs['val_' + l] = o

                callbacks.on_epoch_end(epoch, epoch_logs)
                epoch += 1
                if callback_model.stop_training:
                    break

        finally:
            # if enqueuer is not None:
            #     enqueuer.stop()
            pass

        callbacks.on_train_end()
        return self.history



import keras.backend as K
import tensorflow as tf
import math


def get_custom_objects():
    """
    Functions that returns a dict with all relevant loss functions in this file.
    """
    custom_objects = {
        'loss_direction': loss_direction,
        'loss_uncertainty_mse': loss_uncertainty_mse,
        'loss_uncertainty_mae': loss_uncertainty_mae,
        'loss_uncertainty_gaussian_likelihood': loss_uncertainty_gaussian_likelihood,
        'loss_uncertainty_gaussian_likelihood_dir': loss_uncertainty_gaussian_likelihood_dir,
        'loss_mean_relative_error_energy': loss_mean_relative_error_energy
    }
    return custom_objects


def loss_mean_relative_error_energy(y_true, y_pred):
    """
    Loss function that calculates the mean relative error.
    y_true & y_pred are expected to be e_true & e_pred.
    L = (e_reco - e_true) / e_true

    Returns
    -------
    mre : Mean relative (energy) error loss

    """
    mre = K.abs(y_pred - y_true)/y_true
    return mre


def loss_uncertainty_mae(y_true, y_pred):
    """
    Mean absolute error loss for the uncertainty estimation.
    L = sigma_pred / abs(label_true - label_reco).

    Returns
    -------
    loss : Mean absolute error for uncertainty estimations.

    """
    # order in y_pred: 1) pred label 2) pred label error
    # prevent that the gradient flows back over the label network:
    y_pred_label = K.stop_gradient(y_pred[:, 0])
    y_pred_label_std = y_pred[:, 1]
    y_true_label = y_true[:, 0]

    # (s - |y_true - y_pred|)
    loss = K.abs(y_pred_label_std - K.abs(y_true_label - y_pred_label))
    return loss


def loss_uncertainty_mse(y_true, y_pred):
    """
    Mean squared error loss for the uncertainty estimation.
    L = sigma_pred / abs(label_true - label_reco)**2.

    Returns
    -------
    loss : Mean squared error for uncertainty estimations.

    """
    # order in y_pred: 1) pred label 2) pred label error
    # prevent that the gradient flows back over the label network:
    y_pred_label = K.stop_gradient(y_pred[:, 0])
    y_pred_label_std = y_pred[:, 1]
    y_true_label = y_true[:, 0]

    # (s - |y_true - y_pred|)**2
    loss = K.pow(y_pred_label_std - K.abs(y_true_label - y_pred_label), 2)
    return loss


def loss_uncertainty_gaussian_likelihood(y_true, y_pred):
    """
    Loss function that calculates something similar to a Gaussian Likelihood.
    Requires that y_pred contains only one predicted value (label).
    y_true & y_pred are expected to contain the predicted/true label and
    the predicted std for the label.
    L = ln(std ** 2) + (y_label_pred - y_label_true) / (std ** 2)

    Returns
    -------
    loss : Gaussian Likelihood loss

    """
    # order in y_pred: 1) pred label 2) pred label error
    # prevent that the gradient flows back over the label network:
    y_pred_label = K.stop_gradient(y_pred[:, 0])
    y_pred_label_std = y_pred[:, 1]
    y_true_label = y_true[:, 0]

    # equal to a lower std limit of 3.16 e-2
    eps = tf.constant(1e-3, dtype="float32")
    # y_pred_label_std += eps

    loss = K.log(K.pow(y_pred_label_std, 2) + eps) + K.pow(y_pred_label - y_true_label, 2) / (K.pow(y_pred_label_std, 2) + eps)
    return loss


def loss_uncertainty_gaussian_likelihood_dir(y_true, y_pred):
    """
    Loss function that calculates something similar to a Gaussian
    Likelihood for predicted directions. Requires that y_pred contains
    three predicted values (labels): dir_x, dir_y, dir_z.
    y_true & y_pred are expected to contain the predicted/true label and
    the predicted std for the label.
    L = ln(std ** 2) + (y_label_pred - y_label_true) / (std ** 2)

    Returns
    -------
    loss : Gaussian Likelihood loss for the directional error

    """
    # order in y_pred: 1) pred label 2) pred label error
    # prevent that the gradient flows back over the label network
    y_pred_dir_x, y_pred_dir_y, y_pred_dir_z = K.stop_gradient(y_pred[:, 0]), K.stop_gradient(y_pred[:, 1]), K.stop_gradient(y_pred[:, 2])
    y_pred_std_dir_x, y_pred_std_dir_y, y_pred_std_dir_z = y_pred[:, 3], y_pred[:, 4], y_pred[:, 5]
    y_true_dir_x, y_true_dir_y, y_true_dir_z = y_true[:, 0], y_true[:, 1], y_true[:, 2]

    # equal to a lower std limit of 1e-3
    eps = tf.constant(1e-6, dtype="float32")

    loss_dir_x = K.log(K.pow(y_pred_std_dir_x, 2) + eps) + K.pow(y_pred_dir_x - y_true_dir_x, 2) / (K.pow(y_pred_std_dir_x, 2) + eps)
    loss_dir_y = K.log(K.pow(y_pred_std_dir_y, 2) + eps) + K.pow(y_pred_dir_y - y_true_dir_y, 2) / (K.pow(y_pred_std_dir_y, 2) + eps)
    loss_dir_z = K.log(K.pow(y_pred_std_dir_z, 2) + eps) + K.pow(y_pred_dir_z - y_true_dir_z, 2) / (K.pow(y_pred_std_dir_z, 2) + eps)

    loss = loss_dir_x + loss_dir_y + loss_dir_z
    return loss


def loss_direction(y_true, y_pred):
    """
    Loss function that calculates the space angle between the predicted
    and the true direction.
    Not used anymore, can lead to inf gradients due to tf.acos(space_angle_inner_value)!
    Converts cartesian dirs to spherical coordinate system and then
    calculates the space angle between the two vectors.

    Returns
    -------
    total_loss : Space angle loss
    """
    # define dir_preds
    y_pred = tf.Print(y_pred, [y_pred], message='y_pred', summarize=300)
    y_pred_x, y_pred_y, y_pred_z = y_pred[:, 0], y_pred[:, 1], y_pred[:, 2]
    y_true_x, y_true_y, y_true_z = y_true[:, 0], y_true[:, 1], y_true[:, 2]

    # convert cartesian coordinates to spherical coordinates
    r_pred = K.sqrt(K.pow(y_pred_x, 2) + K.pow(y_pred_y, 2) + K.pow(y_pred_z, 2))
    # y_true input should be normalized! can also use 1 instead of this:
    r_true = K.sqrt(K.pow(y_true_x, 2) + K.pow(y_true_y, 2) + K.pow(y_true_z, 2))

    eps = tf.constant(1e-7, dtype="float32")  # TODO test

    # alternatively zenith_pred = tf.acos(y_pred_z / r_pred + K.epsilon()):
    zenith_pred = tf.atan2(y_pred_z + eps, K.sqrt(K.pow(y_pred_x, 2) + K.pow(y_pred_y, 2)) + eps)
    zenith_true = tf.atan2(y_true_z + eps, K.sqrt(K.pow(y_true_x, 2) + K.pow(y_true_y, 2)) + eps)
    azimuth_pred, azimuth_true = tf.atan2(y_pred_y + eps, y_pred_x + eps), tf.atan2(y_true_y + eps, y_true_x + eps)

    # shift azimuth and zenith by pi / pi/2 in order to make the space angle formula work
    pi = math.pi
    zenith_pred, zenith_true = zenith_pred + tf.constant(pi/float(2), dtype="float32"), zenith_true + tf.constant(pi/float(2), dtype="float32")
    azimuth_pred, azimuth_true = azimuth_pred + tf.constant(pi, dtype="float32"), azimuth_true + tf.constant(pi, dtype="float32")

    # calculate space angle between the two vectors (true, pred) in spherical coordinates, cf. bachelor thesis shallmann
    # protect space angle from acos values outside of [-1,1] range
    space_angle_inner_value = tf.sin(zenith_true) * tf.sin(zenith_pred) * tf.cos(azimuth_true - azimuth_pred) + tf.cos(zenith_true) * tf.cos(zenith_pred)
    space_angle_inner_value = K.clip(space_angle_inner_value, -1, 1)

    space_angle = tf.acos(space_angle_inner_value)
    loss_r = K.abs(r_true - r_pred)

    total_loss = loss_r + space_angle
    # total_loss = tf.Print(total_loss, [total_loss], message='total_loss', summarize=64)
    return total_loss


def loss_direction_grad():
    """
    Similar loss function as in loss_direction, but here it is just used
    for manually calculating the gradient in the  main function. Done in
    order to protect the loss_direction function from inf gradients.

    """
    # define dir_preds
    y_true = np.array([[1, 0, 0], [0, 0, 1]])
    y_true = tf.constant(y_true, shape=y_true.shape, dtype='float32')
    y_pred = np.array([[-1, 0, 0], [0, 0, -1]])
    y_pred = tf.Variable(y_pred, dtype='float32')
    y_pred_x, y_pred_y, y_pred_z = y_pred[:, 0], y_pred[:, 1], y_pred[:, 2]
    y_true_x, y_true_y, y_true_z = y_true[:, 0], y_true[:, 1], y_true[:, 2]

    eps = tf.constant(1e-7, dtype="float32")  # TODO test

    # alternatively zenith_pred = tf.acos(y_pred_z / r_pred + K.epsilon()):
    zenith_pred = tf.atan2(y_pred_z + eps, K.sqrt(K.pow(y_pred_x, 2) + K.pow(y_pred_y, 2)) + eps)
    zenith_true = tf.atan2(y_true_z + eps, K.sqrt(K.pow(y_true_x, 2) + K.pow(y_true_y, 2)) + eps)
    azimuth_pred, azimuth_true = tf.atan2(y_pred_y + eps, y_pred_x + eps), tf.atan2(y_true_y + eps, y_true_x + eps)

    # shift azimuth and zenith by pi / pi/2 in order to make the space angle formula work
    pi = math.pi
    zenith_pred, zenith_true = zenith_pred + tf.constant(pi/float(2), dtype="float32"), zenith_true + tf.constant(pi/float(2), dtype="float32")
    azimuth_pred, azimuth_true = azimuth_pred + tf.constant(pi, dtype="float32"), azimuth_true + tf.constant(pi, dtype="float32")

    # calculate space angle between the two vectors (true, pred) in spherical coordinates, cf. bachelor thesis shallmann
    # protect space angle from acos values outside of [-1,1] range
    space_angle_inner_value = tf.sin(zenith_true) * tf.sin(zenith_pred) * tf.cos(azimuth_true - azimuth_pred) + tf.cos(zenith_true) * tf.cos(zenith_pred)
    space_angle_inner_value = K.clip(space_angle_inner_value, -1, 1)
    space_angle = tf.acos(space_angle_inner_value)

    return space_angle, y_pred


def mean_absolute_error(y_true, y_pred):
    """
    Copy of the Keras mean absolute error function for testing purposes.
    """
    # y_pred = tf.Print(y_pred, [y_pred], message='y_pred', summarize=5)
    # y_true = tf.Print(y_true, [y_true], message='y_true', summarize=5)
    absolute = K.abs(y_pred - y_true)
    # absolute = tf.Print(absolute, [absolute], message='absolute', summarize=5)
    mae = K.mean(absolute, axis=-1)
    return mae


if __name__ == "__main__":
    import numpy as np
    with tf.Session() as sess:
        dir_y_true = np.array([[0.4, 0.1, 0.5], [0.2, -0.4, 0.4], [1, 1, 1], [0.4, 0.2, 0.2], [0.1, 0.2, 0.5]])
        dir_y_pred = np.array([[0.35, 0.05, 0.6], [0.15, -0.3, 0.55], [-1, -1, -1], [-0.4, -0.2, -0.2], [0, 0, 0]])
        x = tf.constant(dir_y_true, shape=dir_y_true.shape, dtype="float32")
        y = tf.constant(dir_y_pred, shape=dir_y_pred.shape, dtype="float32")
        # print('Final loss dir: ' + str(sess.run(loss_direction(x,y))))
        print('Final loss dir: ' + str(sess.run(mean_absolute_error(x, y))))

        # calc gradients for loss_direction
        space_angle_loss_func, y_pred_p = loss_direction_grad()
        init = tf.global_variables_initializer()
        sess.run(init)
        print('Final loss gradients dir: ' + str(sess.run(tf.gradients(space_angle_loss_func, y_pred_p))))

        # errors
        energy_y_true = np.array([[30], [6.35]])
        energy_and_std_y_pred = np.array([[25, 5], [5.84, 0.5]])
        x = tf.constant(energy_y_true, shape=energy_y_true.shape, dtype="float32")
        y = tf.constant(energy_and_std_y_pred, shape=energy_and_std_y_pred.shape, dtype="float32")
        print('Final loss energy uncertainty: ' + str(sess.run(loss_uncertainty_gaussian_likelihood(x, y))))

        dir_y_true = np.array([[0.4, 0.1, 0.5], [0.2, -0.4, 0.4], [1, 1, 1]])
        dir_and_std_y_pred = np.array([[0.35, 0.05, 0.6, 0.05, 0.05, 0.05], [0.15, -0.3, 0.55, 0.05, 0.1, 0.15], [-1, -1, -1, 2, 2, 2]])
        x = tf.constant(dir_y_true, shape=dir_y_true.shape, dtype="float32")
        y = tf.constant(dir_and_std_y_pred, shape=dir_and_std_y_pred.shape, dtype="float32")
        print('Final loss dir uncertainty: ' + str(sess.run(loss_uncertainty_gaussian_likelihood_dir(x, y))))



from keras.layers import Input, Dense, Dropout, Activation, Convolution3D, \
    BatchNormalization, MaxPooling3D, Convolution2D, MaxPooling2D
from keras import backend as K


class ConvBlock:
    """
    2D/3D Convolutional block followed by BatchNorm, Activation,
    MaxPooling and/or Dropout.

    Attributes
    ----------
    conv_dim : int
        Specifies the dimension of the convolutional block, 2D/3D.
    filters : int
        Number of filters used for the convolutional layer.
    kernel_size : int
        Kernel size which is used for all three dimensions.
    pool_size : None or tuple
        Specifies if a MaxPooling layer should be added. e.g. (1,1,2)
        -> sizes for a 3D conv block.
    dropout : float or None
        Adds a dropout layer if the value is not None.
    activation : str or None
        Type of activation function that should be used. E.g. 'linear',
        'relu', 'elu', 'selu'.
    kernel_reg : str or None
        If L2 regularization with 1e-4 should be employed. 'l2' to enable
        the regularization.
    batchnorm : bool
        Adds a batch normalization layer.
    kernel_initializer : string
        Initializer for the kernel weights.

    """
    def __init__(self, conv_dim,
                 filters,
                 kernel_size=3,
                 pool_size=None,
                 dropout=None,
                 activation='relu',
                 kernel_reg=None,
                 batchnorm=False,
                 kernel_initializer="he_normal"):
        self.conv_dim = conv_dim
        self.filters = filters
        self.kernel_size = kernel_size
        self.pool_size = pool_size
        self.dropout = dropout
        self.activation = activation
        self.kernel_reg = kernel_reg
        self.batchnorm = batchnorm
        self.kernel_initializer = kernel_initializer

    def __call__(self, inputs):
        if self.conv_dim == 2:
            convolution_nd = Convolution2D
            max_pooling_nd = MaxPooling2D
        elif self.conv_dim == 3:
            convolution_nd = Convolution3D
            max_pooling_nd = MaxPooling3D
        else:
            raise ValueError('dim must be equal to 2 or 3.')

        x = convolution_nd(self.filters,
                           self.kernel_size,
                           padding='same',
                           kernel_initializer=self.kernel_initializer,
                           use_bias=False,
                           kernel_regularizer=self.kernel_reg)(inputs)

        if self.batchnorm:
            channel_axis = 1 if K.image_data_format() == "channels_first" else -1
            x = BatchNormalization(axis=channel_axis)(x)
        if self.activation is not None:
            x = Activation(self.activation)(x)
        if self.pool_size is not None:
            x = max_pooling_nd(pool_size=self.pool_size, padding='valid')(x)
        if self.dropout is not None:
            x = Dropout(self.dropout)(x)

        return x


class DenseBlock:
    """
    Dense layer followed by BatchNorm, Activation and/or Dropout.

    Attributes
    ----------
    units : int
        Number of neurons of the dense layer.
    dropout : float or None
        Adds a dropout layer if the value is not None.
    activation : str or None
        Type of activation function that should be used. E.g. 'linear',
        'relu', 'elu', 'selu'.
    kernel_reg : str or None
        If L2 regularization with 1e-4 should be employed. 'l2' to enable
        the regularization.
    batchnorm : bool
        Adds a batch normalization layer.

    """
    def __init__(self, units,
                 dropout=None,
                 activation='relu',
                 kernel_reg=None,
                 batchnorm=False,
                 kernel_initializer="he_normal"):
        self.units = units
        self.dropout = dropout
        self.activation = activation
        self.kernel_reg = kernel_reg
        self.batchnorm = batchnorm
        self.kernel_initializer = kernel_initializer

    def __call__(self, inputs):
        x = Dense(self.units, kernel_initializer=self.kernel_initializer,
                  kernel_regularizer=self.kernel_reg)(inputs)
        if self.batchnorm:
            channel_axis = 1 if K.image_data_format() == "channels_first" else -1
            x = BatchNormalization(axis=channel_axis)(x)
        if self.activation is not None:
            x = Activation(self.activation)(x)
        if self.dropout is not None:
            x = Dropout(self.dropout)(x)
        return x


def input_block(input_shapes):
    """
    Build input layers according to a dict mapping the layer names to shapes.

    Parameters
    ----------
    input_shapes : dict
        Keys: Input layer names.
        Values: Their shapes.

    Returns
    -------
    inputs : List
        A list of named keras input layers.

    """
    inputs = []
    for input_name, input_shape in input_shapes.items():
        inputs.append(Input(shape=input_shape, name=input_name, dtype=K.floatx()))
    return inputs
